/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#include "PositionControl.h"
#include "stepper.h"
#include "language.h"
#include "printer.h"
#include "parameter.h"
#include "planner.h"

// constants

XYZ_CONSTS_FROM_CONFIG(signed char, home_dir,  HOME_DIR);
XYZ_CONSTS_FROM_CONFIG(float, home_retract_mm, HOME_RETRACT_MM);
XYZ_CONSTS_FROM_CONFIG(float, max_length,      MAX_LENGTH);
XYZ_CONSTS_FROM_CONFIG(float, base_home_pos,   HOME_POS);
XYZ_CONSTS_FROM_CONFIG(float, base_min_pos,    MIN_POS);
XYZ_CONSTS_FROM_CONFIG(float, base_max_pos,    MAX_POS);

Position_Control::Position_Control()
{

	feedrate = 1500.0;
	saved_feedrate = 0;
	saved_feedmultiply =100;

}

Position_Control::~Position_Control()
{
	// TODO Auto-generated destructor stub
}

void Position_Control::home(bool x_axis, bool y_axis, bool z_axis)
{
	saved_feedrate = feedrate;
	saved_feedmultiply = feedmultiply;
	feedmultiply = 100;
	//   previous_millis_cmd = millis();

	motor.Enable_Endstops(true);

	for(int8_t i=0; i < NUM_AXIS; i++) {
		destination[i] = current_position_mm[i];
	}
	feedrate = 0.0;

	if(x_axis && y_axis )  //first diagonal move
	{
		current_position_mm[X_AXIS] = current_position_mm[Y_AXIS] = 0;

		plan.Set_Position_mm(current_position_mm[X_AXIS], current_position_mm[Y_AXIS], current_position_mm[Z_AXIS], current_position_mm[E_AXIS]);
		destination[X_AXIS] = 1.5 * max_length(X_AXIS) * home_dir(X_AXIS);
		destination[Y_AXIS] = 1.5 * max_length(Y_AXIS) * home_dir(Y_AXIS);
		feedrate = homing_feedrate[X_AXIS];
		if(homing_feedrate[Y_AXIS]<feedrate)
			feedrate = homing_feedrate[Y_AXIS];
		if (max_length(X_AXIS) > max_length(Y_AXIS)) {
			feedrate *= sqrt(pow(max_length(Y_AXIS) / max_length(X_AXIS), 2) + 1);
		} else {
			feedrate *= sqrt(pow(max_length(X_AXIS) / max_length(Y_AXIS), 2) + 1);
		}
		plan.Buffer_Line(destination[X_AXIS], destination[Y_AXIS], destination[Z_AXIS], destination[E_AXIS], feedrate/60, active_extruder);
		motor.Synchronize();

		Set_Axis_At_Home(X_AXIS);
		Set_Axis_At_Home(Y_AXIS);
		plan.Set_Position_mm(current_position_mm[X_AXIS], current_position_mm[Y_AXIS], current_position_mm[Z_AXIS], current_position_mm[E_AXIS]);
		destination[X_AXIS] = current_position_mm[X_AXIS];
		destination[Y_AXIS] = current_position_mm[Y_AXIS];
		plan.Buffer_Line(destination[X_AXIS], destination[Y_AXIS], destination[Z_AXIS], destination[E_AXIS], feedrate/60, active_extruder);
		feedrate = 0.0;
		motor.Synchronize();
		motor.Endstops_hit_on_Purpose();

		current_position_mm[X_AXIS] = destination[X_AXIS];
		current_position_mm[Y_AXIS] = destination[Y_AXIS];
		current_position_mm[Z_AXIS] = destination[Z_AXIS];
	}

	if(x_axis)
	{
		home_axis(X_AXIS);
	}

	if(y_axis)
	{
		home_axis(Y_AXIS);
	}


#if Z_HOME_DIR < 0                      // If homing towards BED do Z last

	if(z_axis)
	{
		home_axis(Z_AXIS);
	}

#endif

	plan.Set_Position_mm(current_position_mm[X_AXIS], current_position_mm[Y_AXIS], current_position_mm[Z_AXIS], current_position_mm[E_AXIS]);

#ifdef ENDSTOPS_ONLY_FOR_HOMING
	motor.Enable_Endstops(false);
#endif

	feedrate = saved_feedrate;
	feedmultiply = saved_feedmultiply;
	//      previous_millis_cmd = millis();
	motor.Endstops_hit_on_Purpose();
}

void Position_Control::home_axis(AxisEnum axis)
{
	current_position_mm[axis] = 0;
	plan.Set_Position_mm(current_position_mm[X_AXIS], current_position_mm[Y_AXIS], current_position_mm[Z_AXIS], current_position_mm[E_AXIS]);

	destination[axis] = 1.5 * max_length(axis) * home_dir(axis);
	feedrate = homing_feedrate[axis];
	plan.Buffer_Line(destination[X_AXIS], destination[Y_AXIS], destination[Z_AXIS], destination[E_AXIS], feedrate/60, active_extruder);
	motor.Synchronize();

	current_position_mm[axis] = 0;
	plan.Set_Position_mm(current_position_mm[X_AXIS], current_position_mm[Y_AXIS], current_position_mm[Z_AXIS], current_position_mm[E_AXIS]);
	destination[axis] = -home_retract_mm(axis) * home_dir(axis);
	plan.Buffer_Line(destination[X_AXIS], destination[Y_AXIS], destination[Z_AXIS], destination[E_AXIS], feedrate/60, active_extruder);
	motor.Synchronize();

	destination[axis] = 2*home_retract_mm(axis) * home_dir(axis);

	feedrate = homing_feedrate[axis]/2 ;

	plan.Buffer_Line(destination[X_AXIS], destination[Y_AXIS], destination[Z_AXIS], destination[E_AXIS], feedrate/60, active_extruder);
	motor.Synchronize();

	Set_Axis_At_Home(axis);
	destination[axis] = current_position_mm[axis];
	feedrate = 0.0;
	motor.Endstops_hit_on_Purpose();
	axis_known_position[axis] = true;
}

bool Position_Control::Set_Tool(uint8_t tool_nr, float move_rate)
{
	/**
	 * T0-T3: Switch tool, usually switching extruders
	 *
	 *   F[mm/min] Set the movement feedrate
	 */
	if (tool_nr >= EXTRUDERS)
	{
		my_printer.f_message(T(MSG_Invalid_Extruder_Num),tool_nr);
		return false;
	}
	else
	{
#if EXTRUDERS > 1
		bool make_move = true;

		motor.Synchronize();

		if (tool_nr != active_extruder) {
			// Save current position to return to after applying extruder offset
			memcpy(destination, current_position_mm, sizeof(destination));

			feedrate = move_rate;
			// Offset extruder (only by XY)
			for (int i = X_AXIS; i <= Y_AXIS; i++)
				current_position_mm[i] += config.extruder_offset[i][tool_nr]
																 - config.extruder_offset[i][active_extruder];
			// Set the new active extruder and position
			active_extruder = tool_nr;

			plan.Set_Position_mm(current_position_mm[X_AXIS], current_position_mm[Y_AXIS], current_position_mm[Z_AXIS], current_position_mm[E_AXIS]);

			// Move to the old position if 'F' was in the parameters
			if (make_move && my_printer.Stopped == false)
				Prepare_Move();

				saved_E_position [active_extruder] = current_position_mm[E_AXIS];
				current_position_mm[E_AXIS] = saved_E_position [tool_nr];
				//the planner position (in steps) needs to be updated too
				plan.Set_Position_mm (E_AXIS, current_position_mm[E_AXIS]);
		}
#endif // EXTRUDERS > 1
		my_printer.f_message(T(MSG_Active_Extruder), active_extruder);
		return true;
	}
}


void Position_Control::Set_Feedrate(float rate)
{
	feedrate = rate;
}

float Position_Control::Feedrate()
{
	return feedrate;
}

void Position_Control::Set_Axis_At_Home(AxisEnum axis)
{
	  current_position_mm[axis] = base_home_pos(axis) + config.add_homeing[axis] ;
	  min_pos[axis] =          base_min_pos(axis) + config.add_homeing[axis];
	  max_pos[axis] =          base_max_pos(axis) + config.add_homeing[axis];
#if EXTRUDERS > 1
	  if (axis == X_AXIS || axis == Y_AXIS)
	  {
		  current_position_mm[axis] += config.extruder_offset[axis][active_extruder] ;
	  }
#endif
}


void Position_Control::Prepare_Move()
{
  clamp_to_software_endstops(destination);

  my_printer.previous_millis_cmd = millis();
#ifdef DELTA
  float difference[NUM_AXIS];
  for (int8_t i=0; i < NUM_AXIS; i++) {
    difference[i] = destination[i] - current_position_mm[i];
  }
  float cartesian_mm = sqrt(sq(difference[X_AXIS]) +
                            sq(difference[Y_AXIS]) +
                            sq(difference[Z_AXIS]));
  if (cartesian_mm < 0.000001) { cartesian_mm = abs(difference[E_AXIS]); }
  if (cartesian_mm < 0.000001) { return; }
  float seconds = 6000 * cartesian_mm / feedrate / feedmultiply;
  int steps = max(1, int(delta_segments_per_second * seconds));
  // SERIAL_ECHOPGM("mm="); SERIAL_ECHO(cartesian_mm);
  // SERIAL_ECHOPGM(" seconds="); SERIAL_ECHO(seconds);
  // SERIAL_ECHOPGM(" steps="); SERIAL_ECHOLN(steps);
  for (int s = 1; s <= steps; s++) {
    float fraction = float(s) / float(steps);
    for(int8_t i=0; i < NUM_AXIS; i++) {
      destination[i] = current_position_mm[i] + difference[i] * fraction;
    }
    calculate_delta(destination);
    plan_buffer_line(delta[X_AXIS], delta[Y_AXIS], delta[Z_AXIS],
                     destination[E_AXIS], feedrate*feedmultiply/60/100.0,
                     active_extruder);
  }
#else

#ifdef DUAL_X_CARRIAGE
  if (active_extruder_parked)
  {
    if (dual_x_carriage_mode == DXC_DUPLICATION_MODE && active_extruder == 0)
    {
      // move duplicate extruder into correct duplication position.
      plan_set_position(inactive_extruder_x_pos, current_position_mm[Y_AXIS], current_position_mm[Z_AXIS], current_position_mm[E_AXIS]);
      plan_buffer_line(current_position_mm[X_AXIS] + duplicate_extruder_x_offset, current_position_mm[Y_AXIS], current_position_mm[Z_AXIS],
          current_position_mm[E_AXIS], max_feedrate[X_AXIS], 1);
      plan_set_position(current_position_mm[X_AXIS], current_position_mm[Y_AXIS], current_position_mm[Z_AXIS], current_position_mm[E_AXIS]);
      st_synchronize();
      extruder_duplication_enabled = true;
      active_extruder_parked = false;
    }
    else if (dual_x_carriage_mode == DXC_AUTO_PARK_MODE) // handle unparking of head
    {
      if (current_position[E_AXIS] == destination[E_AXIS])
      {
        // this is a travel move - skit it but keep track of current position (so that it can later
        // be used as start of first non-travel move)
        if (delayed_move_time != 0xFFFFFFFFUL)
        {
          memcpy(current_position, destination, sizeof(current_position));
          if (destination[Z_AXIS] > raised_parked_position[Z_AXIS])
            raised_parked_position[Z_AXIS] = destination[Z_AXIS];
          delayed_move_time = millis();
          return;
        }
      }
      delayed_move_time = 0;
      // unpark extruder: 1) raise, 2) move into starting XY position, 3) lower
      plan_buffer_line(raised_parked_position[X_AXIS], raised_parked_position[Y_AXIS], raised_parked_position[Z_AXIS],    current_position[E_AXIS], max_feedrate[Z_AXIS], active_extruder);
      plan_buffer_line(current_position[X_AXIS], current_position[Y_AXIS], raised_parked_position[Z_AXIS],
          current_position[E_AXIS], min(max_feedrate[X_AXIS],max_feedrate[Y_AXIS]), active_extruder);
      plan_buffer_line(current_position[X_AXIS], current_position[Y_AXIS], current_position[Z_AXIS],
          current_position[E_AXIS], max_feedrate[Z_AXIS], active_extruder);
      active_extruder_parked = false;
    }
  }
#endif //DUAL_X_CARRIAGE

  // Do not use feedmultiply for E or Z only moves
  if( (Position.current_position_mm[X_AXIS] == destination [X_AXIS]) && (Position.current_position_mm[Y_AXIS] == destination [Y_AXIS])) {
      plan.Buffer_Line(destination[X_AXIS], destination[Y_AXIS], destination[Z_AXIS], destination[E_AXIS], Position.Feedrate()/60, active_extruder);
  }
  else {
    plan.Buffer_Line(destination[X_AXIS], destination[Y_AXIS], destination[Z_AXIS], destination[E_AXIS], Position.Feedrate()*feedmultiply/60/100.0, active_extruder);
  }
#endif //else DELTA
  for(int8_t i=0; i < NUM_AXIS; i++) {
	  Position.current_position_mm[i] = destination[i];
  }
}

void Position_Control::Set_Position_mm(AxisEnum axis, float value_mm)
{
	   current_position_mm[axis] = value_mm;
  plan.Set_Position_mm(axis, value_mm);
}
