/*
    K8400 firmware modification based on original Version provided by Vellemann.

 Copyright (C) 2017 Holger Höhnemann

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef PRINTER_H_
#define PRINTER_H_

#include "messages.h"
#include "MarlinSerial.h"
#include "ultralcd.h"
#include "Configuration_adv.h"
#include "MarlinHOH.h"
#include "stopwatch.h"
#include "PositionControl.h"

#if defined(DUAL_X_CARRIAGE) && defined(X_ENABLE_PIN) && X_ENABLE_PIN > -1 \
    && defined(X2_ENABLE_PIN) && X2_ENABLE_PIN > -1
  #define  enable_x() do { WRITE(X_ENABLE_PIN, X_ENABLE_ON); WRITE(X2_ENABLE_PIN, X_ENABLE_ON); } while (0)
  #define disable_x() do { WRITE(X_ENABLE_PIN,!X_ENABLE_ON); WRITE(X2_ENABLE_PIN,!X_ENABLE_ON); axis_known_position[X_AXIS] = false; } while (0)
#elif defined(X_ENABLE_PIN) && X_ENABLE_PIN > -1
  #define  enable_X() WRITE(X_ENABLE_PIN, X_ENABLE_ON)
  #define disable_X() { WRITE(X_ENABLE_PIN,!X_ENABLE_ON); Position.axis_known_position[X_AXIS] = false; }
#else
  #define enable_x() ;
  #define disable_x() ;
#endif

#if defined(Y_ENABLE_PIN) && Y_ENABLE_PIN > -1
  #ifdef Y_DUAL_STEPPER_DRIVERS
    #define  enable_y() { WRITE(Y_ENABLE_PIN, Y_ENABLE_ON); WRITE(Y2_ENABLE_PIN,  Y_ENABLE_ON); }
    #define disable_y() { WRITE(Y_ENABLE_PIN,!Y_ENABLE_ON); WRITE(Y2_ENABLE_PIN, !Y_ENABLE_ON); my_printer.axis_known_position[Y_AXIS] = false; }
  #else
    #define  enable_Y() WRITE(Y_ENABLE_PIN, Y_ENABLE_ON)
    #define disable_Y() { WRITE(Y_ENABLE_PIN,!Y_ENABLE_ON); Position.axis_known_position[Y_AXIS] = false; }
  #endif
#else
  #define enable_y() ;
  #define disable_y() ;
#endif

#if defined(Z_ENABLE_PIN) && Z_ENABLE_PIN > -1
  #ifdef Z_DUAL_STEPPER_DRIVERS
    #define  enable_z() { WRITE(Z_ENABLE_PIN, Z_ENABLE_ON); WRITE(Z2_ENABLE_PIN, Z_ENABLE_ON); }
    #define disable_z() { WRITE(Z_ENABLE_PIN,!Z_ENABLE_ON); WRITE(Z2_ENABLE_PIN,!Z_ENABLE_ON); Position.axis_known_position[Z_AXIS] = false; }
  #else
    #define  enable_Z() WRITE(Z_ENABLE_PIN, Z_ENABLE_ON)
    #define disable_Z() { WRITE(Z_ENABLE_PIN,!Z_ENABLE_ON); Position.axis_known_position[Z_AXIS] = false; }
  #endif
#else
  #define enable_z() ;
  #define disable_z() ;
#endif

#if defined(E0_ENABLE_PIN) && (E0_ENABLE_PIN > -1)
  #define enable_E0() WRITE(E0_ENABLE_PIN, E_ENABLE_ON)
  #define disable_E0() WRITE(E0_ENABLE_PIN,!E_ENABLE_ON)
#else
  #define enable_e0()  /* nothing */
  #define disable_e0() /* nothing */
#endif

#if (EXTRUDERS > 1) && defined(E1_ENABLE_PIN) && (E1_ENABLE_PIN > -1)
  #define enable_E1() WRITE(E1_ENABLE_PIN, E_ENABLE_ON)
  #define disable_E1() WRITE(E1_ENABLE_PIN,!E_ENABLE_ON)
#else
  #define enable_E1()  /* nothing */
  #define disable_E1() /* nothing */
#endif

#if (EXTRUDERS > 2) && defined(E2_ENABLE_PIN) && (E2_ENABLE_PIN > -1)
  #define enable_e2() WRITE(E2_ENABLE_PIN, E_ENABLE_ON)
  #define disable_e2() WRITE(E2_ENABLE_PIN,!E_ENABLE_ON)
#else
  #define enable_E2()  /* nothing */
  #define disable_E2() /* nothing */
#endif

enum Printerstate {Welcome , Heating, Heating_done, Printing, Waiting, Finished, Message, Error};

class printer: public messages
{
public:
	printer();
	virtual ~printer();

	void Init();
	void message(const char *str);
	void message(char *str);
	void message_ln(const char *str);
	void message_ln(char *str);
	void f_message (const char * formatstr, ...) ;

	void Go_Idle();
	void Reset_Activity_Timeout (unsigned long delay = 10000);
	void Reset_Status_Timeout (unsigned long delay = LCD_TIMEOUT_TO_STATUS);
	void Reset_Fan_Timeout (unsigned long delay = CONTROLLERFAN_SECS*1000UL);
	void Reset_MotorTimeout (unsigned long delay = DEFAULT_STEPPER_DEACTIVE_TIME*1000l);
	void Reset_LCD_Timeout (unsigned long delay = 100);
	void Reset_Error_Timeout(unsigned long delay = 10000);


	inline bool Activity_Timeout() {return millis()>activity_timeout;};
	inline bool Status_Timeout() { return millis() > status_timeout;};
	inline bool Motor_Timeout() { return millis() > stepper_timeout;};
	inline bool Fan_Timeout() {return millis() > fan_timeout;};
	inline bool LCD_Timeout() {return millis() > lcd_timeout;};
	inline bool Error_Timeout() {return millis() > error_timeout;};

	//Inactivity shutdown variables
	unsigned long previous_millis_cmd = 0;
	unsigned long max_inactive_time = 0;
	unsigned long stepper_inactive_time = DEFAULT_STEPPER_DEACTIVE_TIME*1000l;
	unsigned long lastMotor = 0; //Save the time for when a motor was turned on last
	unsigned long lastMotorCheck = 0;
	unsigned long status_timeout = 0;
	unsigned long activity_timeout = 0;
	unsigned long stepper_timeout = 0;
	unsigned long fan_timeout = 0;
	unsigned long lcd_timeout = 0;
	unsigned long error_timeout = 0;

	// Print Job Timer
	Stopwatch print_job_timer;

	bool Stopped=false;

	bool blinker;
	enum Printerstate Status;
	char buffer[70];
private:
	unsigned long delta_t;

	char format[10];
};

extern class printer my_printer;

#endif /* PRINTER_H_ */
